@@ -1,3 +1,4 @@
+#define PING() fprintf (stderr, "%d:%s:%d\n",getpid(),__FILE__,__LINE__),fflush(stderr)
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
 /*
  *  bonobo-activation: A library for accessing bonobo-activation-server.
@@ -253,6 +254,7 @@ bonobo_activation_register_active_server (const char   *iid,
  * 
  * Return value: status of the registration.
  **/
+static volatile int dontstop=0;
 Bonobo_RegistrationResult
 bonobo_activation_register_active_server_ext (const char               *iid,
                                               CORBA_Object              obj,
@@ -277,17 +279,24 @@ bonobo_activation_register_active_server_ext (const char               *iid,
                    CORBA_Object_non_existent (obj, &ev) ? " (nonexistent)" : "");
 #endif
 
-	actid = bonobo_activation_iid_get ();
+	actid = bonobo_activation_iid_get ();PING();
 
         if (actid && strcmp (actid, iid) == 0 && bonobo_activation_private) {
-                retval = Bonobo_ACTIVATION_REG_SUCCESS;
+                retval = Bonobo_ACTIVATION_REG_SUCCESS;PING();
         } else {
 		Bonobo_ActivationEnvironment environment;
-
+PING(); 
+if (getenv ("DEBUG_FUCK_YES"))
+        dontstop=1;
+while (dontstop)
+         ;
+PING();
+if (getenv ("DEBUG_FUCK_YES"))
+        G_BREAKPOINT();
                 od = bonobo_activation_object_directory_get (
                         bonobo_activation_username_get (),
                         bonobo_activation_hostname_get ());
-                
+PING();
                 if (CORBA_Object_is_nil (od, &ev))
                         return Bonobo_ACTIVATION_REG_ERROR;
 
@@ -298,14 +307,25 @@ bonobo_activation_register_active_server_ext (const char               *iid,
                         (od, iid, reg_env ? &environment : &global_reg_env,
                          obj, flags, description ? description : "",
                          existing, bonobo_activation_client_get (), &ev);
-
+PING();
                 if (ev._major == CORBA_SYSTEM_EXCEPTION &&
-                    !strcmp (ev._id, ex_CORBA_BAD_OPERATION)) /* fall-back */
+                    !strcmp (ev._id, ex_CORBA_BAD_OPERATION))
+                {
+PING();
+                            /* fall-back */
+                        CORBA_exception_free (&ev);
                         retval = Bonobo_ObjectDirectory_register_new
                                 (od, iid, reg_env ? &environment : &global_reg_env,
                                  obj, flags, description ? description : "",
                                  existing, &ev);
-
+                }
+PING();
+                if (ev._major != CORBA_NO_EXCEPTION) {
+                        retval = Bonobo_ACTIVATION_REG_ERROR;
+                        g_warning ("Strange exception (%s) from active server registration",
+                                   ev._id);
+                }
+PING();
 		if (reg_env)
 			CORBA_free (environment._buffer);
         }
@@ -322,7 +342,7 @@ bonobo_activation_register_active_server_ext (const char               *iid,
 		int iorfd = bonobo_activation_ior_fd_get ();
 
 		need_ior_printout = FALSE;
-
+PING(); fprintf (stderr, "iorfd=%d\n", iorfd);
 		if (iorfd == 1)
 			fh = stdout;
 		else {
@@ -330,17 +350,17 @@ bonobo_activation_register_active_server_ext (const char               *iid,
 			if (!fh)
 				fh = stdout;
 		}
-
+PING();
 		iorstr = CORBA_ORB_object_to_string (
                         bonobo_activation_orb_get (), obj, &ev);
-
+PING();
 		if (ev._major == CORBA_NO_EXCEPTION) {
 			fprintf (fh, "%s\n", iorstr);
 			CORBA_free (iorstr);
 		}
-
+PING();
 		if (fh != stdout) {
-			fclose (fh);
+PING();			fclose (fh);
 		} else if (iorfd > 2) {
 			close (iorfd);
                 }
